"""Routines for Poetry."""

import re
from pathlib import Path
from typing import List, NamedTuple, Sequence, Tuple

import tomlkit
import tomlkit.container
import tomlkit.items

from .pypi import get_latest_version
from .update import FileContent, Update


def update_pyproject(filename: str) -> Tuple[Sequence[Update], FileContent]:
    """Check updates."""
    updates: List[Update] = []
    config, content = analyze_pyproject(filename)
    for p in config.packages:
        version = re.sub(r"^(?:\^|~|>=)", "", p.version)
        new_version = get_latest_version(p.name)
        if version != new_version:
            updates.append(
                Update(p.name, p.version, new_version, filename, p.category, None)
            )
    return (tuple(updates), content)


class PoetryConfigPackage(NamedTuple):
    """Package in pre-commit configurations."""

    name: str
    version: str
    category: str


class PoetryConfigStruct(NamedTuple):
    """Poetry configuration."""

    filename: str
    packages: Sequence[PoetryConfigPackage]


def analyze_pyproject(
    filename: str,
) -> Tuple[PoetryConfigStruct, FileContent]:
    """Analyze the given pre-commit config file."""
    # It seems that there is no TOML library that tracks line numbers...
    packages: List[PoetryConfigPackage] = []
    text = Path(filename).read_text()
    lines = tuple(text.splitlines())
    data = tomlkit.parse(text)

    if (
        "tool" in data
        and isinstance(
            data["tool"], (tomlkit.items.Table, tomlkit.container.OutOfOrderTableProxy)
        )
        and "poetry" in data["tool"]
        and isinstance(data["tool"]["poetry"], tomlkit.items.Table)
        and "dependencies" in data["tool"]["poetry"]
        and isinstance(data["tool"]["poetry"]["dependencies"], tomlkit.items.Table)
    ):
        deps = data["tool"]["poetry"]["dependencies"]
        packages += _parse_deps(deps, "tool.poetry.dependencies")

    if (
        "tool" in data
        and isinstance(
            data["tool"], (tomlkit.items.Table, tomlkit.container.OutOfOrderTableProxy)
        )
        and "poetry" in data["tool"]
        and isinstance(data["tool"]["poetry"], tomlkit.items.Table)
        and "dev-dependencies" in data["tool"]["poetry"]
        and isinstance(data["tool"]["poetry"]["dev-dependencies"], tomlkit.items.Table)
    ):
        deps = data["tool"]["poetry"]["dev-dependencies"]
        packages += _parse_deps(deps, "tool.poetry.dev-dependencies")

    return PoetryConfigStruct(filename, tuple(packages)), FileContent(filename, lines)


def _parse_deps(deps: tomlkit.items.Table, category: str) -> List[PoetryConfigPackage]:
    packages = []
    for d, v in deps.items():
        d = d.strip()
        if d.lower() == "python":
            continue
        if not isinstance(v, tomlkit.items.String):
            continue
        name = d
        version = str(v).strip()
        m = re.match(r"^((?:\^|~|>=)?)\s*(.*)$", version)
        if m:
            version = m.group(1) + m.group(2)
            package = PoetryConfigPackage(name, version, category)
            packages.append(package)
    return packages



"""Versioning."""

__all__ = ("__version__",)

try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:
    import importlib_metadata  # type: ignore[import,no-redef]

try:
    __version__ = importlib_metadata.version("py-check-updates")
except importlib_metadata.PackageNotFoundError:
    # Undefined during installation.
    __version__ = None  # type: ignore[assignment]



"""A Python dependency update checker."""

from . import poetry, pre_commit, pypi
from ._version import __version__
from .main import check_updates, main
from .update import FileContent, Update

__all__ = (
    "__version__",
    "check_updates",
    "FileContent",
    "main",
    "poetry",
    "pre_commit",
    "pypi",
    "Update",
)



"""Routines for pypi."""

import functools

import requests


def get_latest_version(name: str) -> str:
    """Return the latest version of the given package on PyPI."""
    return _get_latest_version_impl(name.lower())


@functools.lru_cache(maxsize=None)
def _get_latest_version_impl(name: str) -> str:
    r = requests.get(f"https://pypi.org/pypi/{name}/json")
    return r.json()["info"]["version"]  # type: ignore[no-any-return]



"""Update information."""

from typing import NamedTuple, Optional, Sequence


class FileContent(NamedTuple):
    """File content."""

    name: str
    lines: Sequence[str]


class Update(NamedTuple):
    """Update information."""

    name: str
    version: str
    new_version: str
    filename: str
    category: Optional[str]
    line_number: Optional[int]



"""Main routines."""

import argparse
import re
from pathlib import Path
from typing import Optional, Sequence, Tuple, Union

from . import poetry, pre_commit
from .update import FileContent, Update


def check_updates(
    path: Optional[Union[str, Path, Sequence[Union[str, Path]]]] = None
) -> Sequence[Tuple[Sequence[Update], FileContent]]:
    """Check updates."""
    if path is None:
        path = ["."]
    elif isinstance(path, (str, Path)):
        path = [path]

    exclude_dirs = [
        "__pycache__",
        "_build",
        "build",
        "CVS",
        "dist" "node_modules",
        "target",
        r"\..*",
    ]
    exclude_dirs = [f"^{s}$" for s in exclude_dirs]

    files = []

    def _search_files(f: Path) -> None:
        if f.is_dir():
            if any(re.match(e, f.name) for e in exclude_dirs):
                return
            for f1 in f.glob("*"):
                _search_files(f1)
        elif f.is_file():
            name_lower = f.name.lower()
            if name_lower in ("pyproject.toml", ".pre-commit-config.yaml"):
                files.append(f)

    for p in path:
        if isinstance(p, str):
            p = Path(p)
        _search_files(p)
    files = sorted(set(files))

    return tuple(_check_file(f) for f in files)


def _check_file(file: Union[str, Path]) -> Tuple[Sequence[Update], FileContent]:
    if isinstance(file, str):
        file = Path(file)
    name_lower = file.name.lower()

    if name_lower == "pyproject.toml":
        return poetry.update_pyproject(str(file))

    if name_lower == ".pre-commit-config.yaml":
        return pre_commit.update_pre_commit(str(file))

    return (), FileContent(str(file), file.read_text().splitlines())


def main(args: Optional[Sequence[Union[str, Path]]] = None) -> None:
    """Entry point."""
    parser = argparse.ArgumentParser(
        prog="py-check-updates",
        usage=("%(prog)s [options] [--] [file/dir ...]"),
    )
    parser.add_argument("path", nargs="*", help=argparse.SUPPRESS)
    opts = parser.parse_args(args=[str(f) for f in args] if args is not None else None)

    is_first = True
    for updates, content in check_updates(opts.path if opts.path else None):
        if is_first:
            is_first = False
        else:
            print()

        if not updates:
            print(f"{content.name}: OK")
            continue
        else:
            print(f"{content.name}:")
            print()

        # Calculate column widths.

        w_name = 0
        w_version = 0
        w_new_version = 0
        w_category = 0
        w_line_number = 0

        for u in updates:
            w_name = max(w_name, len(u.name))
            w_version = max(w_version, len(u.version))
            w_new_version = max(w_new_version, len(u.new_version))
            if u.category:
                w_category = max(w_category, len(u.category))
            if u.line_number:
                w_line_number = max(w_line_number, len(str(u.line_number)))
        if w_category:
            w_category += 2
        if w_line_number:
            w_line_number += 5

        # Print the result.

        for u in updates:
            cat = f"[{u.category}]" if u.category else ""
            lineno = f"line:{u.line_number}" if u.line_number else ""
            s = ""
            if w_category:
                s += "  " + _str_with_minwidth(cat, w_category)
            if w_line_number:
                s += "  " + _str_with_minwidth(lineno, w_line_number)
            s += "  " + _str_with_minwidth(u.name, w_name)
            s += "  " + _str_with_minwidth(u.version, w_version)
            s += "  " + _str_with_minwidth(u.new_version, w_new_version)
            print("    " + s.strip())


def _str_with_minwidth(s: str, n: int) -> str:
    return s + " " * (n - len(s))



"""Routines for pre-commit."""

from __future__ import annotations

import functools
import re
from pathlib import Path
from typing import Dict, List, NamedTuple, Optional, Sequence, Tuple

import requests
import yaml

from .pypi import get_latest_version
from .update import FileContent, Update


def update_pre_commit(filename: str) -> Tuple[Sequence[Update], FileContent]:
    """Check updates."""
    updates: List[Update] = []
    config, content = analyze_pre_commit_config(filename)
    for p in config.packages:
        if p.language == "python":
            new_version = get_latest_version(p.name)
        else:
            raise ValueError(
                "language not implemented for additional_dependencies: " f"{p.language}"
            )
        if p.version != new_version:
            updates.append(
                Update(p.name, p.version, new_version, filename, None, p.line_number)
            )

    return (tuple(updates), content)


class PreCommitConfigPackage(NamedTuple):
    """Package in pre-commit configurations."""

    name: str
    version: str
    language: str
    filename: str
    line_number: int


class PreCommitConfigStruct(NamedTuple):
    """Pre-commit configuration."""

    filename: str
    packages: Sequence[PreCommitConfigPackage]


def analyze_pre_commit_config(
    filename: str,
) -> Tuple[PreCommitConfigStruct, FileContent]:
    """Analyze the given pre-commit config file."""
    text = Path(filename).read_text()
    lines = tuple(text.splitlines())
    root_node: yaml.Node = yaml.compose(text)  # type: ignore[no-untyped-call]
    packages: List[PreCommitConfigPackage] = []

    if isinstance(root_node, yaml.MappingNode):
        for root_node_key, root_node_value in root_node.value:
            if (
                isinstance(root_node_key, yaml.ScalarNode)
                and root_node_key.value == "repos"
                and isinstance(root_node_value, yaml.SequenceNode)
            ):
                for repo_node in root_node_value.value:
                    if isinstance(repo_node, yaml.MappingNode):
                        name: Optional[str] = None
                        rev: Optional[str] = None
                        hooks: Optional[List[yaml.Node]] = None
                        for repo_node_key, repo_node_value in repo_node.value:
                            if (
                                isinstance(repo_node_key, yaml.ScalarNode)
                                and repo_node_key.value == "repo"
                                and isinstance(repo_node_value, yaml.ScalarNode)
                            ):
                                name = repo_node_value.value
                            elif (
                                isinstance(repo_node_key, yaml.ScalarNode)
                                and repo_node_key.value == "rev"
                                and isinstance(repo_node_value, yaml.ScalarNode)
                            ):
                                rev = repo_node_value.value
                            elif (
                                isinstance(repo_node_key, yaml.ScalarNode)
                                and repo_node_key.value == "hooks"
                                and isinstance(repo_node_value, yaml.SequenceNode)
                            ):
                                hooks = repo_node_value.value
                        if name and rev and hooks:
                            languages = analyze_pre_commit_hooks(name, rev)
                            for h in hooks:
                                packages += _parse_hook(h, filename, languages)

    return PreCommitConfigStruct(filename, tuple(packages)), FileContent(
        filename, lines
    )


def _parse_hook(
    hook: yaml.Node, filename: str, languages: PreCommitHooksStruct
) -> List[PreCommitConfigPackage]:
    packages = []
    if isinstance(hook, yaml.MappingNode):
        hook_language: Optional[str] = None
        additional_dependencies: Optional[List[yaml.Node]] = None
        for key, value in hook.value:
            if (
                isinstance(key, yaml.ScalarNode)
                and key.value == "id"
                and isinstance(value, yaml.ScalarNode)
            ):
                hook_id = value.value
                hook_language = languages.languages[hook_id]
            elif (
                isinstance(key, yaml.ScalarNode)
                and key.value == "additional_dependencies"
                and isinstance(value, yaml.SequenceNode)
            ):
                additional_dependencies = value.value
        if hook_language and additional_dependencies:
            for dep_node in additional_dependencies:
                if isinstance(dep_node, yaml.ScalarNode):
                    dep_str = dep_node.value
                    if hook_language == "python":
                        m = re.match(r"^(.*)==(.*)$", dep_str)
                        if m:
                            dep_name = m.group(1).strip()
                            dep_ver = m.group(2).strip()
                            package = PreCommitConfigPackage(
                                dep_name,
                                dep_ver,
                                hook_language,
                                filename,
                                dep_node.start_mark.line + 1,
                            )
                            packages.append(package)
                    else:
                        # Ignore other languages for now.
                        pass
    return packages


HookId = str
HookLanguage = str


class PreCommitHooksStruct(NamedTuple):
    """Pre-commit hooks."""

    name: str
    rev: str
    languages: Dict[HookId, HookLanguage]


@functools.lru_cache(maxsize=None)
def analyze_pre_commit_hooks(name: str, rev: str) -> PreCommitHooksStruct:
    """Analyze the given pre-commit hooks on GitHub."""
    languages = {}
    data = yaml.load(_get_pre_commit_hook_impl(name, rev), Loader=yaml.SafeLoader)
    if isinstance(data, list):
        for d in data:
            if isinstance(d, dict):
                if "id" in d and "language" in d:
                    hook_id = d["id"]
                    hook_language = d["language"]
                    if isinstance(hook_id, str) and isinstance(hook_language, str):
                        languages[hook_id] = hook_language

    return PreCommitHooksStruct(name, rev, languages)


def _get_pre_commit_hook_impl(name: str, rev: str) -> str:
    github_prefix = "https://github.com/"
    if name.lower().startswith(github_prefix):
        url = f"https://raw.githubusercontent.com/{name[len(github_prefix):]}/{rev}"
    else:
        return ""

    r = requests.get(f"{url}/.pre-commit-hooks.yaml")
    return r.text



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['check_updates']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0', 'requests>=2.28.1,<3.0.0', 'tomlkit>=0.11.0,<0.12.0']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata>=1.4']}

entry_points = \
{'console_scripts': ['py-check-updates = check_updates:main']}

setup_kwargs = {
    'name': 'py-check-updates',
    'version': '0.0.1rc1',
    'description': 'A Python dependency update checker.',
    'long_description': 'py-check-updates\n================\n\nThis is a dependency update checker to assist Python package developers.\n\nIf you are the kind of paranoid person who checks whether packages\nin your dependency lists are updated once a week\nand you do it manually for some technical reasons,\nthen this program is for you.\n\nCurrently, the program checks updates (it does not automatically update them, though) for:\n\n- Top-level dependencies in `pyproject.toml` for [Poetry](https://python-poetry.org/)\n  (see [poetry#2684](https://github.com/python-poetry/poetry/issues/2684)).\n  Limited to simple versions of the form `package = "(^|~|>=)?version"`.\n\n- Additional dependencies of hooks in `.pre-commit-config.yaml` for [pre-commit](https://pre-commit.com/)\n  (see [pre-commit#1351](https://github.com/pre-commit/pre-commit/issues/1351)).\n  Limited to simple Python dependencies of the form `package==version`.\n\nInstallation\n------------\n\n```bash\npip install py-check-updates\n```\n\nUsage\n-----\n\n```bash\npy-check-updates\n```\n',
    'author': 'Takahiro Ueda',
    'author_email': 'takahiro.ueda@yahoo.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/tueda/py-check-updates',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
